#!/usr/bin/env python3

"""A textual base app with common features like a logging window"""

import asyncio
import logging
from argparse import ArgumentParser
from pathlib import Path

from rich.logging import RichHandler
from rich.markup import escape as markup_escape
from textual import on
from textual.app import App, ComposeResult
from textual.message import Message
from textual.scrollbar import ScrollTo
from textual.widgets import RichLog

from .logging_helper import LOG_LEVELS


def log() -> logging.Logger:
    """Returns the logger instance to use here"""
    return logging.getLogger("trickkiste.base_app")


class RichLogHandler(RichHandler):
    """Redirects rich.RichHanlder capabilities to a textual.RichLog"""

    def __init__(self, widget: RichLog, level: int = logging.INFO):
        super().__init__(show_path=False, markup=True, show_time=False, level=level)
        self.widget: RichLog = widget

    def emit(self, record: logging.LogRecord) -> None:
        record.args = record.args and tuple(
            markup_escape(arg) if isinstance(arg, str) else arg for arg in record.args
        )
        record.msg = markup_escape(record.msg)
        self.widget.write(
            self.render(
                record=record,
                message_renderable=self.render_message(record, self.format(record)),
                traceback=None,
            )
        )


class LockingRichLog(RichLog):
    """A RichLog which turns off autoscroll when scrolling manually"""

    @on(ScrollTo)
    def on_scroll_to(self, _event: Message) -> None:
        """Mandatory comment"""
        self.auto_scroll = self.is_vertical_scroll_end


class TuiBaseApp(App[None]):
    """A nice UI for Sauron stuff"""

    CSS_PATH = Path(__file__).parent / "base_tui_app.css"

    def __init__(self, logger_funcname: bool = True) -> None:
        super().__init__()
        self._richlog = LockingRichLog()
        self._logger_funcname = logger_funcname
        self._log_level: int | str = logging.INFO

    def add_default_arguments(self, parser: ArgumentParser) -> None:
        """Adds arguments to @parser we need in every app"""
        parser.add_argument(
            "--log-level",
            "-l",
            type=str.upper,
            choices=LOG_LEVELS,
            default="INFO",
        )

    def compose(self) -> ComposeResult:
        """Set up the UI"""
        yield self._richlog

    async def on_mount(self) -> None:
        """UI entry point"""
        logging.getLogger().handlers = [handler := RichLogHandler(self._richlog)]
        optional_funcname = "│ [grey53]%(funcName)-32s[/] " if self._logger_funcname else ""
        handler.setFormatter(
            logging.Formatter(
                f"│ %(asctime)s {optional_funcname}│ [bold white]%(message)s[/]",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        )
        self.set_log_level(self._log_level)
        if hasattr(self, "initialize"):
            await self.initialize()

    def execute(self) -> None:
        """Wrapper for async run and optional cleanup if provided"""
        asyncio.run(self.run_async())
        if hasattr(self, "cleanup"):
            self.cleanup()

    def set_log_level(self, level: int | str) -> None:
        """Sets the overall log level for internal log console"""

        # log level for everything
        logging.getLogger().setLevel(logging.DEBUG if level == "ALL_DEBUG" else logging.WARNING)

        # log level for provided @logger
        used_level = getattr(logging, level.split("_")[-1]) if isinstance(level, str) else level

        for handler in logging.getLogger().handlers:
            handler.setLevel(used_level)

        self._log_level = level



#!/usr/bin/env python3

"""Example for fancy logging"""

import logging
from argparse import ArgumentParser

from trickkiste.logging_helper import apply_common_logging_cli_args, setup_logging


def log() -> logging.Logger:
    """Returns the logger instance to use here"""
    return logging.getLogger(__name__)


def main() -> None:
    """Runs this"""
    parser = ArgumentParser(__doc__)
    apply_common_logging_cli_args(parser)
    args = parser.parse_args()

    setup_logging(log(), level=args.log_level)

    log().debug("debug message")
    log().info("info message")
    log().warning("warning message")
    log().error("error message")

    logging.getLogger("other.module").debug("only shown with level=ALL_DEBUG")
    logging.getLogger("other.module").info("only shown with level=ALL_DEBUG")
    logging.getLogger("other.module").warning("this is visible")
    logging.getLogger("other.module").error("this is also visible")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Mixed common stuff not big enough for a separate module"""

import hashlib
import logging
import os
import re
import shlex
from collections.abc import Iterator, Mapping
from contextlib import contextmanager, suppress
from datetime import datetime
from pathlib import Path
from subprocess import DEVNULL, check_output

from dateutil import tz


def log() -> logging.Logger:
    """Returns the logger instance to use here"""
    return logging.getLogger("trickkiste.misc")


def md5from(filepath: Path) -> None | str:
    """Returns an MD5 sum from contents of file provided"""
    with suppress(FileNotFoundError):
        with open(filepath, "rb") as input_file:
            file_hash = hashlib.md5()
            while chunk := input_file.read(1 << 16):
                file_hash.update(chunk)
            return file_hash.hexdigest()
    return None


@contextmanager
def cwd(path: Path) -> Iterator[None]:
    """Changes working directory and returns to previous on exit."""
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


def dur_str(seconds: float, fixed: bool = False) -> str:
    """Turns a duration defined by @seconds into a string like '1d:2h:3m'
    If @fixed is True, numbers will be 0-padded for uniform width.
    Negative values for @seconds are not supported (yet)
    >>> dur_str(42)
    '42s'
    >>> dur_str(12345)
    '3h:25m:45s'
    """
    if not fixed and not seconds:
        return "0s"
    digits = 2 if fixed else 1
    days = f"{int(seconds//86400):0{digits}d}d" if fixed or seconds >= 86400 else ""
    hours = (
        f"{int(seconds//3600%24):0{digits}d}h"
        if fixed or seconds >= 3600 and (seconds % 86400)
        else ""
    )
    minutes = (
        f"{int(seconds//60%60):0{digits}d}m" if fixed or seconds >= 60 and (seconds % 3600) else ""
    )
    seconds_str = (
        f"{int(seconds%60):0{digits}d}s" if not fixed and ((seconds % 60) or seconds == 0) else ""
    )
    return ":".join(e for e in (days, hours, minutes, seconds_str) if e)


def age_str(now: float | datetime, age: None | int | datetime, fixed: bool = False) -> str:
    """Turn a number of seconds into something human readable
    >>> age_str(1700000000, 1600000000)
    '1157d:9h:46m:40s'
    """
    if age is None:
        return "--"
    age_ts = age.timestamp() if isinstance(age, datetime) else age
    now_ts = now.timestamp() if isinstance(now, datetime) else now
    if (age_ts or now_ts) <= 0.0:
        return "--"
    return dur_str(now_ts - age_ts, fixed=fixed)


def date_str(timestamp: int | datetime, datefmt: str = "%Y.%m.%d-%H:%M:%S") -> str:
    """Returns a uniform time string from a timestamp or a datetime
    >>> date_str(datetime.strptime("1980.01.04-12:55:02", "%Y.%m.%d-%H:%M:%S"))
    '1980.01.04-12:55:02'
    """
    if not timestamp:
        return "--"
    date_dt = timestamp if isinstance(timestamp, datetime) else datetime.fromtimestamp(timestamp)
    if date_dt.year < 1000:
        return "--"
    return (date_dt).strftime(datefmt)


def date_from(timestamp: int | float | str) -> None | datetime:
    """
    >>> str(date_from("2023-07-14T15:05:32.174200714+02:00"))
    '2023-07-14 15:05:32+02:00'
    >>> str(date_from("2023-07-24T21:25:26.89389821+02:00"))
    '2023-07-24 21:25:26+02:00'
    """
    try:
        if isinstance(timestamp, datetime):
            return timestamp

        if isinstance(timestamp, (int, float)):
            return datetime.fromtimestamp(timestamp)

        if timestamp[-1] == "Z":
            return (
                datetime.strptime(timestamp[:19], "%Y-%m-%dT%H:%M:%S")
                .replace(tzinfo=tz.tzutc())
                .astimezone(tz.tzlocal())
            )
        if re.match(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+\+\d{2}:\d{2}$", timestamp):
            timestamp = timestamp[:19] + timestamp[-6:]
        return datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S%z")
    except OverflowError:
        return None
    except Exception as exc:  # pylint: disable=broad-except
        raise ValueError(f"Could not parse datetime from <{timestamp!r}> ({exc})") from exc


def parse_age(string: str) -> int:
    """Return seconds extracted from something parsable as an age ('3d:4h')
    >>> parse_age("3600")
    3600
    >>> parse_age("3h")
    10800
    >>> parse_age("1d:10h")
    122400
    >>> parse_age("1d10h")
    122400
    >>> parse_age("1.5d")
    129600
    """
    number = r"\d*(?:\.(?:\d*)?)?"
    with suppress(ValueError):
        return int(string)
    if match := re.match(
        (
            "(?i)^"
            f"(?:({number})d)?"
            "(?::)?"
            f"(?:({number})h)?"
            "(?::)?"
            f"(?:({number})m)?"
            "(?::)?"
            f"(?:({number})s)?$"
        ),
        string,
    ):
        days, hours, mins, secs = match.groups()
        return int(
            float(days or "0") * 86400
            + float(hours or "0") * 3600
            + float(mins or "0") * 60
            + float(secs or "0")
        )
    raise ValueError(f"Cannot turn {string!r} into duration")


def smart_split(string: str, delimiter: str = ",") -> Iterator[str]:
    """Like str.split but takes quotes and parenthesis into account
    >>> list(smart_split("(=,),'a=x',\\"b=y\\""))
    ['(=,)', "'a=x'", '"b=y"']
    """
    splits: list[int] = []
    closers: list[str] = []
    parenthizers = {'"': '"', "'": "'", "[": "]", "(": ")"}
    for i, char in enumerate(string):
        if char == delimiter and not closers:
            splits.append(i)
        elif closers and char == closers[-1]:
            closers.pop()
        elif char in parenthizers:
            closers.append(parenthizers[char])
    start = 0
    for pos in splits:
        yield string[start:pos]
        start = pos + 1
    yield string[start:]


def split_params(string: str) -> Mapping[str, str]:
    """Splits a 'string packed map' into a dict
    >>> split_params("foo=23,bar=42,true='pi=3,14'")
    {'foo': '23', 'bar': '42', 'true': "'pi=3,14'"}
    """
    return {k: v for p in smart_split(string, ",") if p for k, v in (p.split("=", 1),)}


def compact_dict(
    mapping: Mapping[str, float | str], *, maxlen: None | int = 10, delim: str = ", "
) -> str:
    """Turns a dict into a 'string packed map' (for making a dict human readable)
    >>> compact_dict({'foo': '23', 'bar': '42'})
    'foo=23, bar=42'
    """

    def short(string: str) -> str:
        return string if maxlen is None or len(string) <= maxlen else f"{string[:maxlen-2]}.."

    return delim.join(
        f"{k}={short_str}" for k, v in mapping.items() if (short_str := short(str(v)))
    )


def process_output(cmd: str) -> str:
    """Return command output as one blob"""
    return check_output(shlex.split(cmd), stderr=DEVNULL, text=True)



#!/usr/bin/env python3

"""Common stuff shared among modules"""

import logging
import os
import sys
import traceback
from argparse import ArgumentParser
from collections.abc import Iterable

from rich.console import Console
from rich.logging import RichHandler
from rich.markup import escape as markup_escape

LOG_LEVELS = ("ALL_DEBUG", "DEBUG", "INFO", "WARN", "ERROR", "CRITICAL")


def apply_common_logging_cli_args(parser: ArgumentParser) -> None:
    """Decorates given @parser with arguments for logging"""
    parser.add_argument(
        "--log-level",
        "-l",
        choices=LOG_LEVELS,
        help="Sets the logging level - ALL_DEBUG sets all other loggers to DEBUG, too",
        type=str.upper,
        default="INFO",
    )


def stack_str(depth: int = 0) -> str:
    """Returns a short local function call stack"""

    def stack_fns() -> Iterable[str]:
        stack = list(
            reversed(
                traceback.extract_stack(sys._getframe(depth))  # pylint: disable=protected-access
            )
        )

        for site in stack:
            if site.filename != stack[0].filename or site.name == "<module>":
                break
            yield site.name

    return ">".join(reversed(list(stack_fns())))


def setup_logging(logger: logging.Logger | str, level: str | int = "INFO") -> None:
    """Make logging fun"""

    class CustomLogger(  # pylint: disable=unused-variable
        logging.getLoggerClass()  # type: ignore[misc]
    ):
        """Injects the 'stack' element"""

        def makeRecord(self, *args: object, **kwargs: object) -> logging.LogRecord:
            """Adds 'stack' element to given record"""
            kwargs.setdefault("extra", {})["stack"] = stack_str(5)  # type: ignore[index]
            return super().makeRecord(*args, **kwargs)  # type: ignore[no-any-return]

    # currently this introduces more problems then benefits
    # logging.setLoggerClass(CustomLogger)

    # log level for everything
    root_log_level = logging.DEBUG if level == "ALL_DEBUG" else logging.WARNING

    # log level for provided @logger
    used_level = getattr(logging, level.split("_")[-1]) if isinstance(level, str) else level

    if not logging.getLogger().hasHandlers():
        logging.getLogger().setLevel(root_log_level)
        shandler = RichHandler(
            show_time=False,
            show_path=False,
            markup=True,
            console=Console(
                stderr=True, color_system="standard" if os.environ.get("FORCE_COLOR") else "auto"
            ),
        )
        logging.getLogger().addHandler(shandler)
        shandler.setLevel(min(used_level, root_log_level))
        shandler.setFormatter(logging.Formatter("│ [grey]%(name)-15s[/] │ [bold]%(message)s[/]"))

        # logging.basicConfig(
        #   format="%(name)s %(levelname)s: %(message)s",
        #   datefmt="%Y-%m-%d %H:%M:%S",
        #   level=logging.DEBUG if level == "ALL_DEBUG" else logging.WARNING,
        # )

        def markup_escaper(record: logging.LogRecord) -> bool:
            record.args = record.args and tuple(
                markup_escape(arg) if isinstance(arg, str) else arg for arg in record.args
            )
            record.msg = markup_escape(record.msg)
            return True

        shandler.addFilter(markup_escaper)

    # for lev in LOG_LEVELS:
    #    logging.addLevelName(getattr(logging, lev), f"{lev[0] * 2}")

    logging.getLogger("urllib3.connectionpool").setLevel(logging.INFO)
    (logging.getLogger(logger) if isinstance(logger, str) else logger).setLevel(used_level)



#!/usr/bin/env python3

"""Too lazy for a real test - I'm using this example to check if colorful logging works"""

import asyncio
import logging

from textual import on, work
from textual.app import ComposeResult
from textual.widgets import Tree

from trickkiste.base_tui_app import TuiBaseApp


def log() -> logging.Logger:
    """Returns the logger instance to use here"""
    return logging.getLogger("trickkiste.example_tui")


class ExampleTUI(TuiBaseApp):
    """A little example TUI"""

    def __init__(self) -> None:
        super().__init__()
        self.tree_widget: Tree[None] = Tree("A Tree")

    def compose(self) -> ComposeResult:
        """Set up the UI"""
        yield self.tree_widget
        yield from super().compose()

    @on(Tree.NodeSelected)
    def on_node_selected(self, event: Tree.NodeSelected[None]) -> None:
        """React on clicking a node (links handled differently)"""
        log().debug("clicked %s", event.node.label)

    @work(exit_on_error=True)
    async def produce(self) -> None:
        """Busy worker task continuously rebuilding the job tree"""
        log().info("first message from async worker")
        cpu_node = self.tree_widget.root.add("CPU")
        mem_node = self.tree_widget.root.add("MEM")
        disk_node = self.tree_widget.root.add("DISK")
        self.tree_widget.root.expand()
        while True:
            log().info("step..")
            cpu_node.set_label("TBD")
            mem_node.set_label("TBD")
            disk_node.set_label("TBD")
            await asyncio.sleep(5)

    async def on_mount(self) -> None:
        """UI entry point"""
        self.produce()


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.WARNING)
    log().setLevel(logging.DEBUG)
    ExampleTUI().execute()
